"""
Klingon Tools: A collection of utilities for various tasks.

This module provides access to the LogTools class, which offers
logging functionality for the Klingon Tools package.
"""

from .log_tools import LogTools

__version__ = "0.0.0"

__all__ = ["LogTools"]
